from typing import (
    Iterable,
    Tuple,
)

import brotli
from izihawa_types.var import (
    process_varint,
    process_varstr,
)

from .tantivy_coder import TantivyCoder


def auto_process_varint(data):
    var, length = process_varint(data, inverted=True)
    return var, data[length:]


def read_footer(data: bytes) -> Tuple[bytes, bytes]:
    footer_len = int.from_bytes(data[-4:], 'little', signed=False)
    data = data[:-4]
    body, footer = data[:-footer_len], data[-footer_len:]
    return body, footer


def read_store(body: bytes) -> Tuple[bytes, bytes]:
    offset = int.from_bytes(body[-8:], 'little', signed=False)
    return body[:offset], body[offset:-8]


def parse_footer(footer: bytes) -> Tuple[int, int, bytes]:
    footer_length, footer = auto_process_varint(footer)
    version = int.from_bytes(footer[:4], 'little', signed=False)
    crc = int.from_bytes(footer[4:8], 'little', signed=False)
    compression, _ = process_varstr(footer[8:], inverted=True)
    return version, crc, compression


def read_layers(skip_index):
    layers = []
    skip_index_length, skip_index = auto_process_varint(skip_index)
    start_offset = 0
    for i in range(skip_index_length):
        end_offset, skip_index = auto_process_varint(skip_index)
        layers.append((start_offset, end_offset))
        start_offset = end_offset
    return layers, skip_index


def parse_blocks(last_layer):
    checkpoints = []
    while last_layer:
        length, last_layer = auto_process_varint(last_layer)
        if length == 0:
            continue
        doc_id, last_layer = auto_process_varint(last_layer)
        start_offset, last_layer = auto_process_varint(last_layer)
        for i in range(length):
            num_docs, last_layer = auto_process_varint(last_layer)
            block_num_bytes, last_layer = auto_process_varint(last_layer)
            checkpoints.append({
                'start_doc': doc_id,
                'end_doc': doc_id + num_docs,
                'start_offset': start_offset,
                'end_offset': start_offset + block_num_bytes,
            })
            doc_id += num_docs
            start_offset += block_num_bytes
    return checkpoints


class TantivyReader:
    def __init__(self, store: bytes, coder: TantivyCoder):
        self.store = store
        self.coder = coder

    def documents(self) -> Iterable[dict]:
        """
        Iterates over all document inside `.store`

        Returns:
            Document generator
        """
        body, footer = read_footer(self.store)
        store, skip_index = read_store(body)
        layers, skip_index = read_layers(skip_index)

        checkpoints = parse_blocks(skip_index[layers[-1][0]:layers[-1][1]])

        for checkpoint in checkpoints:
            decompressed = brotli.decompress(store[checkpoint['start_offset']:checkpoint['end_offset']])
            while decompressed:
                doc_length, decompressed = auto_process_varint(decompressed)
                document = self.coder.decode_document(decompressed)
                yield document
                decompressed = decompressed[doc_length:]



from typing import (
    Optional,
    Tuple,
    Union,
)

from izihawa_types.var import (
    process_varint,
    process_varstr,
    varint,
    varstr,
)


class TantivyDocument(dict):
    def __getattr__(self, item):
        return self.get(item)

    def __setattr__(self, item, value):
        self[item] = value


class TantivyCoder:
    def __init__(self, schema_config: dict):
        self.schema_config = schema_config.copy()
        if 'multi_fields' in self.schema_config:
            self.schema_config['multi_fields'] = set(self.schema_config['multi_fields'])

    def decode_field(self, reader: bytes) -> Tuple[int, Union[Optional[bytes], int], bytes]:
        field_id = int.from_bytes(reader[:4], 'little', signed=False)
        reader = reader[4:]
        type_ = int.from_bytes(reader[:1], 'little', signed=False)
        reader = reader[1:]
        if type_ == 0:
            value, length = process_varstr(reader, inverted=True)
            value = value.decode()
            reader = reader[length:]
        elif type_ == 1:
            value = int.from_bytes(reader[:8], 'little', signed=False)
            reader = reader[8:]
        elif type_ == 2:
            value = int.from_bytes(reader[:8], 'little', signed=True)
            reader = reader[8:]
        else:
            return field_id, None, reader
        return field_id, value, reader

    def encode_field(self, field_id: int, type_: str, values) -> bytes:
        # ToDo: Unknown types
        buf = bytes()
        for value in values:
            buf += int.to_bytes(field_id, 4, 'little', signed=False)
            if type_ == 'text':
                buf += int.to_bytes(0, 1, 'little', signed=False)
                buf += varstr(value.encode(), inverted=True)
            elif type_ == 'u64':
                buf += int.to_bytes(1, 1, 'little', signed=False)
                buf += int.to_bytes(value, 8, 'little', signed=False)
            elif type_ == 'i64':
                buf += int.to_bytes(2, 1, 'little', signed=False)
                buf += int.to_bytes(value, 8, 'little', signed=True)
        return buf

    def decode_document(self, document: bytes) -> TantivyDocument:
        num_field_values, length = process_varint(document, inverted=True)
        reader = document[length:]
        document = TantivyDocument()
        for _ in range(num_field_values):
            field_id, value, reader = self.decode_field(reader)
            name = self.schema_config['schema'][field_id]['name']
            if name in self.schema_config['multi_fields']:
                if name not in document:
                    document[name] = []
                document[name].append(value)
            elif name not in document:
                document[name] = value
        return document

    def encode_document(self, document: TantivyDocument) -> bytes:
        encoded = bytes()
        count = 0
        for (i, field) in enumerate(self.schema_config['schema']):
            if hasattr(document, field['name']) and getattr(document, field['name']) is not None:
                values = getattr(document, field['name'])
                if field['name'] not in self.schema_config['multi_fields']:
                    values = [values]
                count += len(values)
                encoded += self.encode_field(i, field['type'], values)
        encoded = varint(count, inverted=True) + encoded
        return encoded
